@@ -44,7 +44,6 @@ if sys.platform == "win32":
 
 from meadowrun.run_job import (
     AllocCloudInstance,
-    AllocCloudInstances,
     AwsSecret,
     AzureSecret,
     CondaEnvironmentYmlFile,
@@ -53,22 +52,24 @@ from meadowrun.run_job import (
     ContainerInterpreterBase,
     Deployment,
     InterpreterSpecFile,
+    KubernetesSecret,
     LocalCondaInterpreter,
     LocalInterpreter,
     LocalPipInterpreter,
     PipRequirementsFile,
     PoetryProjectPath,
+    Secret,
     run_command,
     run_function,
     run_map,
 )
-from meadowrun.run_job_local import LocalHost
-from meadowrun.run_job_core import SshHost
+from meadowrun.run_job_local import Host, LocalHost
+from meadowrun.run_job_core import Resources, SshHost
+from meadowrun.kubernetes_integration import Kubernetes
 
 
 __all__ = [
     "AllocCloudInstance",
-    "AllocCloudInstances",
     "AwsSecret",
     "AzureSecret",
     "CondaEnvironmentYmlFile",
@@ -77,14 +78,19 @@ __all__ = [
     "ContainerInterpreterBase",
     "Deployment",
     "InterpreterSpecFile",
+    "KubernetesSecret",
     "LocalCondaInterpreter",
     "LocalInterpreter",
     "LocalPipInterpreter",
     "PipRequirementsFile",
     "PoetryProjectPath",
+    "Secret",
     "run_command",
     "run_function",
     "run_map",
+    "Host",
     "LocalHost",
+    "Resources",
     "SshHost",
+    "Kubernetes",
 ]
