@@ -55,11 +55,11 @@ from meadowrun.instance_allocation import (
     _TInstanceState,
     allocate_jobs_to_instances,
 )
-from meadowrun.instance_selection import Resources, CloudInstance
+from meadowrun.instance_selection import ResourcesInternal, CloudInstance
 
 if TYPE_CHECKING:
     from meadowrun.meadowrun_pb2 import Job
-from meadowrun.run_job_core import AllocCloudInstancesInternal, JobCompletion, SshHost
+from meadowrun.run_job_core import JobCompletion, SshHost
 
 # SEE ALSO ec2_alloc_stub.py
 
@@ -67,26 +67,26 @@ from meadowrun.run_job_core import AllocCloudInstancesInternal, JobCompletion, S
 # replicate into each region.
 _AMIS = {
     "plain": {
-        "us-east-2": "ami-0e842f37f2769ee33",
-        "us-east-1": "ami-01962adb562193031",
-        "us-west-1": "ami-0b4181c5d98c58a25",
-        "us-west-2": "ami-0e6486b71c4a54048",
-        "eu-central-1": "ami-0f4eaecf685167922",
-        "eu-west-1": "ami-02f8d1e7c2db6d9bf",
-        "eu-west-2": "ami-045662b1304e77342",
-        "eu-west-3": "ami-026be7722484138c1",
-        "eu-north-1": "ami-0c0a2686efe7a7a81",
+        "us-east-2": "ami-08766f069e82a723d",
+        "us-east-1": "ami-007175f8ebc9b7403",
+        "us-west-1": "ami-0ff608d06f5151d38",
+        "us-west-2": "ami-0339ed7ea96c87620",
+        "eu-central-1": "ami-096ea2c6825846c78",
+        "eu-west-1": "ami-05f662e3d0efb4cc9",
+        "eu-west-2": "ami-0b41c4ec6f7169a94",
+        "eu-west-3": "ami-0e20d27f93f34352c",
+        "eu-north-1": "ami-0f1b0711aee472818",
     },
     "cuda": {
-        "us-east-2": "ami-0bbe0129d34b9e852",
-        "us-east-1": "ami-08064543b56faebf8",
-        "us-west-1": "ami-0821383b8d71b08c0",
-        "us-west-2": "ami-0974ee249205ca251",
-        "eu-central-1": "ami-0973e1e645697c8f5",
-        "eu-west-1": "ami-09ff36908b735d62a",
-        "eu-west-2": "ami-063562cc36c753b78",
-        "eu-west-3": "ami-00e6164aaf0c0e284",
-        "eu-north-1": "ami-0f7ee24731e5b9627",
+        "us-east-2": "ami-0b9f9011ae18b5d4f",
+        "us-east-1": "ami-0e0769031ea69f297",
+        "us-west-1": "ami-043e3443dba006c3c",
+        "us-west-2": "ami-03f9617f990fae806",
+        "eu-central-1": "ami-0a5182ba83a42ecab",
+        "eu-west-1": "ami-0b994ff5c4534a740",
+        "eu-west-2": "ami-05d686d8a98f6f513",
+        "eu-west-3": "ami-0b2502d2572fe31e8",
+        "eu-north-1": "ami-0e87e280b496c8628",
     },
 }
 SSH_USER = "ubuntu"
@@ -172,8 +172,8 @@ class EC2InstanceRegistrar(InstanceRegistrar[_InstanceState]):
         self,
         public_address: str,
         name: str,
-        resources_available: Resources,
-        running_jobs: List[Tuple[str, Resources]],
+        resources_available: ResourcesInternal,
+        running_jobs: List[Tuple[str, ResourcesInternal]],
     ) -> None:
         # TODO we should probably enforce types on logical_cpu_available and
         # memory_gb_available
@@ -260,7 +260,7 @@ class EC2InstanceRegistrar(InstanceRegistrar[_InstanceState]):
             _InstanceState(
                 item[_PUBLIC_ADDRESS],
                 item[_INSTANCE_ID],
-                Resources.from_decimals(
+                ResourcesInternal.from_decimals(
                     item[_RESOURCES_AVAILABLE], item[_NON_CONSUMABLE_RESOURCES]
                 ),
                 # This is a bit of hack, but for the EC2InstanceRegistrar, we know that
@@ -295,7 +295,7 @@ class EC2InstanceRegistrar(InstanceRegistrar[_InstanceState]):
     async def allocate_jobs_to_instance(
         self,
         instance: _InstanceState,
-        resources_allocated_per_job: Resources,
+        resources_allocated_per_job: ResourcesInternal,
         new_job_ids: List[str],
     ) -> bool:
         if len(new_job_ids) == 0:
@@ -403,29 +403,32 @@ class EC2InstanceRegistrar(InstanceRegistrar[_InstanceState]):
         return True
 
     async def launch_instances(
-        self, instances_spec: AllocCloudInstancesInternal
+        self,
+        resources_required_per_task: ResourcesInternal,
+        num_concurrent_tasks: int,
+        region_name: str,
     ) -> Sequence[CloudInstance]:
-        if "nvidia" in instances_spec.resources_required_per_task.non_consumable:
+        if "nvidia" in resources_required_per_task.non_consumable:
             ami_type = "cuda"
         else:
             ami_type = "plain"
 
-        if instances_spec.region_name not in _AMIS[ami_type]:
+        if region_name not in _AMIS[ami_type]:
             raise ValueError(
-                f"The meadowrun AMI is not available in {instances_spec.region_name}. "
-                "Please ask the meadowrun maintainers to add support for this region: "
+                f"The meadowrun AMI is not available in {region_name}. Please ask the "
+                "meadowrun maintainers to add support for this region: "
                 "https://github.com/meadowdata/meadowrun/issues and try a supported "
                 f"region for now: {', '.join(_AMIS[ami_type].keys())}"
             )
-        ami = _AMIS[ami_type][instances_spec.region_name]
+        ami = _AMIS[ami_type][region_name]
 
         return await launch_ec2_instances(
-            instances_spec.resources_required_per_task,
-            instances_spec.num_concurrent_tasks,
+            resources_required_per_task,
+            num_concurrent_tasks,
             ami,
-            region_name=instances_spec.region_name,
+            region_name=region_name,
             # TODO we should let users add their own security groups
-            security_group_ids=[get_ssh_security_group_id(instances_spec.region_name)],
+            security_group_ids=[get_ssh_security_group_id(region_name)],
             # TODO we should let users set their own IAM role as long as it grants
             # access to the dynamodb table we need for deallocation
             iam_role_name=_EC2_ROLE_INSTANCE_PROFILE,
@@ -466,7 +469,7 @@ class EC2InstanceRegistrar(InstanceRegistrar[_InstanceState]):
 
 async def run_job_ec2_instance_registrar(
     job: Job,
-    resources_required: Resources,
+    resources_required: ResourcesInternal,
     region_name: Optional[str],
 ) -> JobCompletion[Any]:
     """Runs the specified job on EC2. Creates an EC2InstanceRegistrar"""
@@ -476,7 +479,9 @@ async def run_job_ec2_instance_registrar(
     async with EC2InstanceRegistrar(region_name, "create") as instance_registrar:
         hosts = await allocate_jobs_to_instances(
             instance_registrar,
-            AllocCloudInstancesInternal(resources_required, 1, region_name),
+            resources_required,
+            1,
+            region_name,
             job.ports,
         )
 
@@ -490,4 +495,6 @@ async def run_job_ec2_instance_registrar(
     # remains mostly an internal concept
     job.job_id = job_ids[0]
 
-    return await SshHost(host, SSH_USER, pkey, ("EC2", region_name)).run_job(job)
+    return await SshHost(host, SSH_USER, pkey, ("EC2", region_name)).run_job(
+        resources_required, job
+    )
