@@ -3,6 +3,8 @@ from __future__ import annotations
 import asyncio
 import base64
 import functools
+from io import BytesIO
+import itertools
 import os
 import pickle
 import traceback
@@ -19,15 +21,14 @@ from typing import (
     TypeVar,
     cast,
 )
-import itertools
 
 import aiobotocore.session
 import boto3
-
+from meadowrun.aws_integration import s3
 from meadowrun.aws_integration.aws_core import _get_default_region_name
 from meadowrun.aws_integration.ec2_instance_allocation import (
-    EC2InstanceRegistrar,
     SSH_USER,
+    EC2InstanceRegistrar,
 )
 from meadowrun.aws_integration.ec2_ssh_keys import get_meadowrun_ssh_key
 from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
@@ -37,11 +38,11 @@ from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
 from meadowrun.instance_allocation import allocate_jobs_to_instances
 
 if TYPE_CHECKING:
-    from meadowrun.instance_selection import Resources
-from meadowrun.meadowrun_pb2 import GridTask, ProcessState, GridTaskStateResponse
-from meadowrun.run_job_core import RunMapHelper, AllocCloudInstancesInternal
-from meadowrun.shared import pickle_exception
+    from meadowrun.instance_selection import ResourcesInternal
 
+from meadowrun.meadowrun_pb2 import GridTask, GridTaskStateResponse, ProcessState
+from meadowrun.run_job_core import RunMapHelper
+from meadowrun.shared import pickle_exception
 
 _REQUEST_QUEUE_NAME_PREFIX = "meadowrun-task-request-"
 _RESULT_QUEUE_NAME_PREFIX = "meadowrun-task-result-"
@@ -83,10 +84,6 @@ async def _create_queues_for_job(job_id: str, region_name: str) -> Tuple[str, st
     hyphens
     """
 
-    # TODO we should automatically delete these queues also. I think the best way is to
-    # have get_results periodically tag the queue with a timestamp and have a lambda
-    # periodically clean them up
-
     async with aiobotocore.session.get_session().create_client(
         "sqs", region_name=region_name
     ) as client:
@@ -143,22 +140,8 @@ async def _add_tasks(
             # this function can only take 10 messages at a time, so we chunk into
             # batches of 10
             # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Client.send_message_batch
-            result = await client.send_message_batch(
-                QueueUrl=request_queue_url,
-                Entries=[
-                    {
-                        "Id": str(i),
-                        "MessageBody": base64.b64encode(
-                            GridTask(
-                                task_id=i, pickled_function_arguments=pickle.dumps(task)
-                            ).SerializeToString()
-                        ).decode("utf-8"),
-                        # TODO replace 0 with a retry count
-                        "MessageDeduplicationId": f"{i}:0",
-                        "MessageGroupId": "_",
-                    }
-                    for i, task in tasks_chunk
-                ],
+            result = await _send_or_upload_message_batch(
+                client, region_name, request_queue_url, tasks_chunk
             )
             if "Failed" in result:
                 raise ValueError(
@@ -166,6 +149,95 @@ async def _add_tasks(
                 )
 
 
+_BODY_INLINE = "inline:"
+_BODY_STORED = "stored:"
+MAX_MESSAGE_SIZE = 262_144 - 1024  # max size 256 KiB - 1KiB buffer
+
+
+async def _send_or_upload_message_batch(
+    client: Any,
+    region_name: str,
+    queue: str,
+    tasks_chunk: List[Tuple[int, Any]],
+) -> Any:
+    entries = []
+    for i, task in tasks_chunk:
+        message_body_bytes = base64.b64encode(
+            GridTask(
+                task_id=i, pickled_function_arguments=pickle.dumps(task)
+            ).SerializeToString()
+        )
+        if len(message_body_bytes) < MAX_MESSAGE_SIZE:
+            entries.append(
+                {
+                    "Id": str(i),
+                    "MessageBody": _BODY_INLINE + message_body_bytes.decode("utf-8"),
+                    # TODO replace 0 with a retry count
+                    "MessageDeduplicationId": f"{i}:0",
+                    "MessageGroupId": "_",
+                }
+            )
+        else:
+            msg_id = "sqs/" + str(uuid.uuid4())
+            with BytesIO(message_body_bytes) as bs:
+                s3.upload_fileobj(msg_id, bs, region_name)
+            entries.append(
+                {
+                    "Id": str(i),
+                    "MessageBody": _BODY_STORED + msg_id,
+                    # TODO replace 0 with a retry count
+                    "MessageDeduplicationId": f"{i}:0",
+                    "MessageGroupId": "_",
+                }
+            )
+    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Client.send_message
+    return await client.send_message_batch(QueueUrl=queue, Entries=entries)
+
+
+def _send_or_upload_message(
+    client: Any,
+    region_name: str,
+    queue: str,
+    response: GridTaskStateResponse,
+    deduplication_id: str,
+) -> None:
+
+    message_body_bytes = base64.b64encode(response.SerializeToString())
+    if len(message_body_bytes) < MAX_MESSAGE_SIZE:
+        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Client.send_message
+        client.send_message(
+            QueueUrl=queue,
+            MessageBody=_BODY_INLINE + message_body_bytes.decode("utf-8"),
+            # TODO replace 0 here with a retry count
+            MessageDeduplicationId=deduplication_id,
+            MessageGroupId="_",
+        )
+    else:
+        msg_id = "sqs/" + str(uuid.uuid4())
+        with BytesIO(message_body_bytes) as bs:
+            s3.upload_fileobj(msg_id, bs, region_name)
+        client.send_message(
+            QueueUrl=queue,
+            MessageBody=_BODY_STORED + msg_id,
+            # TODO replace 0 here with a retry count
+            MessageDeduplicationId=deduplication_id,
+            MessageGroupId="_",
+        )
+
+
+def _get_result_from_body(body: Any, region_name: str, task_result: Any) -> None:
+    if body.startswith(_BODY_INLINE):
+        task_result.ParseFromString(
+            base64.b64decode(body[len(_BODY_INLINE) :].encode("utf-8"))
+        )
+    elif body.startswith(_BODY_STORED):
+        with BytesIO() as bs:
+            s3.download_fileobj(body[len(_BODY_STORED) :], bs, region_name)
+            task_result.ParseFromString(base64.b64decode(bs.getvalue()))
+    else:
+        raise ValueError(f"Malformed SQS message body: {body}")
+
+
 def _get_task(
     request_queue_url: str,
     result_queue_url: str,
@@ -195,39 +267,34 @@ def _get_task(
     messages = result["Messages"]
     if len(messages) != 1:
         raise ValueError(f"Requested one message but got {len(messages)}")
-    else:
-        # parse the task request
-        task = GridTask()
-        task.ParseFromString(base64.b64decode(messages[0]["Body"].encode("utf-8")))
 
-        # send the RUN_REQUESTED message on the result queue
-        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Client.send_message
-        client.send_message(
-            QueueUrl=result_queue_url,
-            MessageBody=base64.b64encode(
-                GridTaskStateResponse(
-                    task_id=task.task_id,
-                    process_state=ProcessState(
-                        state=ProcessState.ProcessStateEnum.RUN_REQUESTED,
-                        # TODO needs to include public address and worker id
-                    ),
-                ).SerializeToString()
-            ).decode("utf-8"),
-            # TODO replace 0 here with a retry count
-            MessageDeduplicationId=(
-                f"{task.task_id}:0:{public_address}:{worker_id}:requested"
+    # parse the task request
+    task = GridTask()
+    _get_result_from_body(messages[0]["Body"], region_name, task)
+
+    # send the RUN_REQUESTED message on the result queue
+    _send_or_upload_message(
+        client,
+        region_name,
+        result_queue_url,
+        GridTaskStateResponse(
+            task_id=task.task_id,
+            process_state=ProcessState(
+                state=ProcessState.ProcessStateEnum.RUN_REQUESTED,
+                # TODO needs to include public address and worker id
             ),
-            MessageGroupId="_",
-        )
+        ),
+        # TODO replace 0 here with a retry count
+        f"{task.task_id}:0:{public_address}:{worker_id}:requested",
+    )
+    # acknowledge receipt/delete the task request message so we don't have duplicate
+    # tasks running
+    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Client.delete_message
+    client.delete_message(
+        QueueUrl=request_queue_url, ReceiptHandle=messages[0]["ReceiptHandle"]
+    )
 
-        # acknowledge receipt/delete the task request message so we don't have duplicate
-        # tasks running
-        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Client.delete_message
-        client.delete_message(
-            QueueUrl=request_queue_url, ReceiptHandle=messages[0]["ReceiptHandle"]
-        )
-
-        return task
+    return task
 
 
 def _complete_task(
@@ -242,18 +309,14 @@ def _complete_task(
     Sends a message to the result queue that the specified task has completed with the
     specified process_state
     """
-    boto3.client("sqs", region_name=region_name).send_message(
-        QueueUrl=result_queue_url,
-        MessageBody=base64.b64encode(
-            GridTaskStateResponse(
-                task_id=task.task_id, process_state=process_state
-            ).SerializeToString()
-        ).decode("utf-8"),
+    client = boto3.client("sqs", region_name=region_name)
+    _send_or_upload_message(
+        client,
+        region_name,
+        result_queue_url,
+        GridTaskStateResponse(task_id=task.task_id, process_state=process_state),
         # TODO replace 0 here with a retry count
-        MessageDeduplicationId=(
-            f"{task.task_id}:0:{public_address}:{worker_id}:completed"
-        ),
-        MessageGroupId="_",
+        f"{task.task_id}:0:{public_address}:{worker_id}:completed",
     )
 
 
@@ -365,9 +428,7 @@ async def get_results(
                 for message in receive_result["Messages"]:
                     receipt_handles.append(message["ReceiptHandle"])
                     task_result = GridTaskStateResponse()
-                    task_result.ParseFromString(
-                        base64.b64decode(message["Body"].encode("utf-8"))
-                    )
+                    _get_result_from_body(message["Body"], region_name, task_result)
 
                     if (
                         task_result.process_state.state
@@ -375,6 +436,7 @@ async def get_results(
                     ):
                         running_tasks[task_result.task_id] = task_result.process_state
                     elif task_results[task_result.task_id] is None:
+                        running_tasks[task_result.task_id] = None
                         task_results[task_result.task_id] = task_result.process_state
                         task_results_received += 1
 
@@ -414,7 +476,7 @@ async def prepare_ec2_run_map(
     function: Callable[[_T], _U],
     tasks: Sequence[_T],
     region_name: Optional[str],
-    resources_required_per_task: Resources,
+    resources_required_per_task: ResourcesInternal,
     num_concurrent_tasks: int,
     ports: Optional[Sequence[str]],
 ) -> RunMapHelper:
@@ -432,11 +494,9 @@ async def prepare_ec2_run_map(
     async with EC2InstanceRegistrar(region_name, "create") as instance_registrar:
         allocated_hosts = await allocate_jobs_to_instances(
             instance_registrar,
-            AllocCloudInstancesInternal(
-                resources_required_per_task,
-                num_concurrent_tasks,
-                region_name,
-            ),
+            resources_required_per_task,
+            num_concurrent_tasks,
+            region_name,
             ports,
         )
 
