@@ -1,5 +1,5 @@
 import hashlib
-from typing import Dict, List, Optional, Tuple
+from typing import BinaryIO, Dict, List, Optional, Tuple
 
 import boto3
 import boto3.exceptions
@@ -49,12 +49,18 @@ def ensure_bucket(
 
     bucket_name = _get_bucket_name(region_name)
 
-    location = {"LocationConstraint": region_name}
+    # us-east-1 cannot be specified as a LocationConstraint because it is the default
+    # region
+    # https://stackoverflow.com/questions/51912072/invalidlocationconstraint-error-while-creating-s3-bucket-when-the-used-command-i
+    if region_name == "us-east-1":
+        additional_parameters = {}
+    else:
+        additional_parameters = {
+            "CreateBucketConfiguration": {"LocationConstraint": region_name}
+        }
 
     success, _ = ignore_boto3_error_code(
-        lambda: s3.create_bucket(
-            Bucket=bucket_name, CreateBucketConfiguration=location
-        ),
+        lambda: s3.create_bucket(Bucket=bucket_name, **additional_parameters),
         "BucketAlreadyOwnedByYou",
     )
     s3.put_bucket_lifecycle_configuration(
@@ -129,6 +135,18 @@ async def download_file(
     s3.download_file(bucket_name, object_name, file_name)
 
 
+def upload_fileobj(object_name: str, file_obj: BinaryIO, region_name: str) -> None:
+    s3 = boto3.client("s3", region_name=region_name)
+    bucket_name = _get_bucket_name(region_name)
+    s3.upload_fileobj(file_obj, bucket_name, object_name)
+
+
+def download_fileobj(object_name: str, file_obj: BinaryIO, region_name: str) -> None:
+    s3 = boto3.client("s3", region_name=region_name)
+    bucket_name = _get_bucket_name(region_name)
+    s3.download_fileobj(bucket_name, object_name, file_obj)
+
+
 def delete_bucket(region_name: str) -> None:
     """Deletes the meadowrun bucket"""
     s3 = boto3.resource("s3", region_name=region_name)
