@@ -16,11 +16,14 @@ async def _pip_freeze(python_interpreter: str) -> str:
     # and in a "regular" virtualenv. We should keep an eye on that issue, though, as it
     # looks like this difference in behavior was not totally intentional and may change
     # in the future.
+    # Update Aug 22: we try to find pyproject.toml and poetry.lock before getting here,
+    # so we're using pip freeze now. Direct URLs are needed to support packages
+    # installed via git URLs.
     env = os.environ.copy()
     env["PIP_DISABLE_PIP_VERSION_CHECK"] = "1"
     p = await asyncio.create_subprocess_exec(
         python_interpreter,
-        *("-m", "pip", "list", "--format=freeze"),
+        *("-m", "pip", "freeze"),
         stdout=asyncio.subprocess.PIPE,
         stderr=asyncio.subprocess.PIPE,
         env=env,
