@@ -3,7 +3,9 @@ from __future__ import annotations
 import abc
 import asyncio
 import dataclasses
+import os
 import os.path
+import pathlib
 import pickle
 import platform
 import shlex
@@ -58,14 +60,15 @@ if TYPE_CHECKING:
         VersionedCodeDeployment,
         VersionedInterpreterDeployment,
     )
+    from meadowrun.instance_selection import ResourcesInternal
 from meadowrun.docker_controller import get_registry_domain
-from meadowrun.instance_selection import Resources
 from meadowrun.meadowrun_pb2 import (
     AwsSecretProto,
     AzureSecretProto,
     CodeZipFile,
     ContainerAtDigest,
     ContainerAtTag,
+    ContainerImage,
     Credentials,
     CredentialsSourceMessage,
     EnvironmentSpec,
@@ -74,6 +77,7 @@ from meadowrun.meadowrun_pb2 import (
     GitRepoBranch,
     GitRepoCommit,
     Job,
+    KubernetesSecretProto,
     PyCommandJob,
     PyFunctionJob,
     QualifiedFunctionName,
@@ -87,7 +91,13 @@ from meadowrun.pip_integration import (
     pip_freeze_without_local_current_interpreter,
     pip_freeze_without_local_other_interpreter,
 )
-from meadowrun.run_job_core import CloudProviderType, Host, JobCompletion, SshHost
+from meadowrun.run_job_core import (
+    CloudProviderType,
+    Host,
+    JobCompletion,
+    Resources,
+    SshHost,
+)
 
 _T = TypeVar("_T")
 _U = TypeVar("_U")
@@ -172,6 +182,32 @@ class AzureSecret(Secret):
         )
 
 
+@dataclasses.dataclass
+class KubernetesSecret(Secret):
+    """
+    A Kubernetes secret
+
+    Attributes:
+        secret_name: The name of the secret
+    """
+
+    secret_name: str
+
+    def _to_credentials_source(
+        self,
+        service: CredentialsService,
+        service_url: str,
+        credentials_type: Credentials.Type.ValueType,
+    ) -> CredentialsSourceForService:
+        return CredentialsSourceForService(
+            service=service,
+            service_url=service_url,
+            source=KubernetesSecretProto(
+                credentials_type=credentials_type, secret_name=self.secret_name
+            ),
+        )
+
+
 class ContainerInterpreterBase(abc.ABC):
     """An abstract base class for specifying a container as a python interpreter"""
 
@@ -375,13 +411,25 @@ class LocalCurrentInterpreter(LocalInterpreter):
 
         # next, check if we're in a poetry environment. Unfortunately it doesn't seem
         # like there's a way to detect that we're in a poetry environment unless the
-        # current working directory contains the pyproject.toml and poetry.lock files.
-        # If for some reason this isn't the case, we'll fall through to the pip-based
-        # case, which will mostly work for poetry environments.
-        if os.path.isfile("pyproject.toml") and os.path.isfile("poetry.lock"):
-            with open("pyproject.toml", encoding="utf-8") as project_file:
+        # current working directory or a parent contains the pyproject.toml and
+        # poetry.lock files. If for some reason this isn't the case, we'll fall through
+        # to the pip-based case.
+        cwd = pathlib.Path.cwd()
+        candidates = [cwd]
+        candidates.extend(cwd.parents)
+        is_poetry_env = False
+        for path in candidates:
+            pyproject_file = path / "pyproject.toml"
+            poetry_lock_file = path / "poetry.lock"
+
+            if pyproject_file.exists() and poetry_lock_file.exists():
+                is_poetry_env = True
+                break
+
+        if is_poetry_env:
+            with open(pyproject_file, encoding="utf-8") as project_file:
                 project_file_contents = project_file.read()
-            with open("poetry.lock", encoding="utf-8") as lock_file:
+            with open(poetry_lock_file, encoding="utf-8") as lock_file:
                 lock_file_contents = lock_file.read()
             print("Mirroring current poetry environment")
             return EnvironmentSpec(
@@ -481,10 +529,11 @@ class Deployment:
     async def mirror_local(
         cls,
         include_sys_path: bool = True,
-        additional_paths: Iterable[str] = tuple(),
+        additional_python_paths: Iterable[str] = tuple(),
         interpreter: Union[
             LocalInterpreter, InterpreterSpecFile, ContainerInterpreterBase, None
         ] = None,
+        working_directory_globs: Union[str, Iterable[str], None] = None,
         environment_variables: Optional[Dict[str, str]] = None,
     ) -> Deployment:
         """A deployment that mirrors the local environment and code.
@@ -493,7 +542,8 @@ class Deployment:
             include_sys_path: if True, find python code in the paths in sys.path
                 (effectively "your local code"), copies them to the remote machines, and
                 add them to sys.path there. Ignores any installed packages.
-            additional_paths: local code paths to copy to the remote machine.
+            additional_python_paths: local code paths to copy and add to python path on
+                the remote machine.
             interpreter: Specifies the environment/interpreter to use. Defaults to None
                 which will detect the currently activated env. Alternatively, you can
                 explicitly specify a locally installed python environment with
@@ -502,6 +552,18 @@ class Deployment:
                 [CondaEnvironmentYmlFile][meadowrun.CondaEnvironmentYmlFile],
                 [PipRequirementsFile][meadowrun.PipRequirementsFile],
                 [PoetryProjectPath][meadowrun.PoetryProjectPath]
+            working_directory_globs: Most of the time, your current working directory
+                will be on `sys.path`, and so any *.py files from your current working
+                directory will be uploaded by `mirror_local`. However, any other types
+                of files will be ignored by default. You can specify files to include
+                from your current working directory with this argument. Examples:
+                `"foo/bar.txt"` will include the specified file. `"*.txt"` will specify
+                txt files in your current directory (but not recursively). `"**/*.txt"`
+                will specify all txt files in your current directory recursively (e.g.
+                will capture both `1.txt` and `foo/2.txt`). `"foo/**/*.txt"` will
+                capture all txt files in the foo directory. If you specify files like
+                this, you will be able to access them with relative paths in the remote
+                code the same way you reference them locally.
             environment_variables: e.g. `{"PYTHONHASHSEED": "0"}`. These environment
                 variables will be set in the remote environment.
 
@@ -573,15 +635,23 @@ class Deployment:
         else:
             raise ValueError(f"Unexpected type of interpreter {type(interpreter)}")
 
+        if working_directory_globs is None:
+            working_directory_globs = ()
+        elif isinstance(working_directory_globs, str):
+            working_directory_globs = [working_directory_globs]
+
         # annoyingly, this tmp dir now gets deleted in run_local when the file
         # has been uploaded/unpacked depending on the Host implementation
         tmp_dir = tempfile.mkdtemp()
-        zip_file_path, zip_paths = local_code.zip(
-            tmp_dir, include_sys_path, additional_paths
+        zip_file_path, zip_python_paths, zip_cwd = local_code.zip_local_code(
+            tmp_dir,
+            include_sys_path,
+            additional_python_paths,
+            working_directory_globs=working_directory_globs,
         )
 
         url = urllib.parse.urlunparse(("file", "", zip_file_path, "", "", ""))
-        code = CodeZipFile(url=url, code_paths=zip_paths)
+        code = CodeZipFile(url=url, code_paths=zip_python_paths, cwd_path=zip_cwd)
 
         return cls(interpreter_spec, code, environment_variables, credentials_sources)
 
@@ -702,6 +772,76 @@ class Deployment:
 
         return cls(interpreter_spec, code, environment_variables, credentials_sources)
 
+    @classmethod
+    def container_image(
+        cls,
+        repository: str,
+        tag: str = "latest",
+        username_password_secret: Optional[Secret] = None,
+        environment_variables: Optional[Dict[str, str]] = None,
+    ) -> Deployment:
+        """
+        A deployment based on a docker container image
+
+        Arguments:
+            repository: The name of the docker container image repository, e.g. `python`
+                or `quay.io/minio/minio`.
+            tag: Combined with repository, will be used like `{repository}:{tag}`.
+                Defaults to `latest`
+            environment_variables: e.g. `{"PYTHONHASHSEED": "0"}`. These environment
+                variables will be set in the remote environment.
+        """
+        credentials_sources = []
+        if username_password_secret is not None:
+            credentials_sources.append(
+                username_password_secret._to_credentials_source(
+                    "DOCKER",
+                    get_registry_domain(repository)[0],
+                    Credentials.Type.USERNAME_PASSWORD,
+                )
+            )
+        return cls(
+            ContainerAtTag(repository=repository, tag=tag),
+            ServerAvailableFolder(),
+            environment_variables,
+            credentials_sources,
+        )
+
+    @classmethod
+    def container_image_at_digest(
+        cls,
+        repository: str,
+        digest: str,
+        username_password_secret: Optional[Secret] = None,
+        environment_variables: Optional[Dict[str, str]] = None,
+    ) -> Deployment:
+        """
+        A deployment based on a docker container image
+
+        Arguments:
+            repository: The name of the docker container image repository, e.g. `python`
+                or `quay.io/minio/minio`.
+            digest: Combined with repository, will be used like `{repository}@{tag}`.
+                Defaults to `latest` if digest is not specified.
+            environment_variables: e.g. `{"PYTHONHASHSEED": "0"}`. These environment
+                variables will be set in the remote environment.
+        """
+        credentials_sources = []
+        if username_password_secret is not None:
+            credentials_sources.append(
+                username_password_secret._to_credentials_source(
+                    "DOCKER",
+                    get_registry_domain(repository)[0],
+                    Credentials.Type.USERNAME_PASSWORD,
+                )
+            )
+        return cls(
+            ContainerAtDigest(repository=repository, digest=digest),
+            ServerAvailableFolder(),
+            environment_variables,
+            credentials_sources,
+        )
+
 
 def _credentials_source_message(
     credentials_source: CredentialsSourceForService,
@@ -716,6 +856,8 @@ def _credentials_source_message(
         result.azure_secret.CopyFrom(credentials_source.source)
     elif isinstance(credentials_source.source, ServerAvailableFile):
         result.server_available_file.CopyFrom(credentials_source.source)
+    elif isinstance(credentials_source.source, KubernetesSecretProto):
+        result.kubernetes_secret.CopyFrom(credentials_source.source)
     else:
         raise ValueError(
             f"Unknown type of credentials source {type(credentials_source.source)}"
@@ -784,10 +926,6 @@ class _DeploymentTarget(Enum):
         else:
             return _DeploymentTarget.LOCAL
 
-    @staticmethod
-    def from_hosts(hosts: AllocCloudInstances) -> _DeploymentTarget:
-        return _DeploymentTarget.from_cloud_provider(hosts.cloud_provider)
-
 
 async def _prepare_code_deployment(
     code_deploy: Union[CodeDeployment, VersionedCodeDeployment],
@@ -828,57 +966,34 @@ async def _prepare_code_deployment(
         raise ValueError(f"Unexpected value for target {target}")
 
 
-def _needs_cuda(flags_required: Union[Iterable[str], str, None]) -> bool:
-    if flags_required is None:
-        return False
-    if isinstance(flags_required, str):
-        return "nvidia" == flags_required
-    return "nvidia" in flags_required
-
-
 @dataclasses.dataclass(frozen=True)
 class AllocCloudInstance(Host):
     """
-    Specifies the requirements for a cloud instance (e.g. an EC2 instance or Azure VM).
+    Specifies that the job should be run on a dynamically allocated cloud instance (i.e.
+    either an EC2 instance or Azure VM) depending on the cloud_provider parameter. Any
+    existing Meadowrun-managed instances will be reused if available. If none are
+    available, Meadowrun will launch the cheapest instance that meets the resource
+    requirements for a job.
+
+    resources_required must be provided with the AllocCloudInstance Host.
 
     Attributes:
-        logical_cpu_required:
-        memory_gb_required:
-        interruption_probability_threshold: Specifies what interruption probability
-            percent is acceptable. E.g. `80` means that any instance type with an
-            interruption probability less than 80% can be used. Use `0` to indicate that
-            only on-demand instance are acceptable (i.e. do not use spot instances)
         cloud_provider: `EC2` or `AzureVM`
-        gpus_required:
-        gpu_memory_required: Total GPU memory required across all GPUs
-        flags_required: E.g. "intel", "avx512", etc.
         region_name:
     """
 
-    logical_cpu_required: float
-    memory_gb_required: float
-    interruption_probability_threshold: float
     cloud_provider: CloudProviderType
-    gpus_required: Optional[float] = None
-    gpu_memory_required: Optional[float] = None
-    flags_required: Union[Iterable[str], str, None] = None
     region_name: Optional[str] = None
 
-    def uses_gpu(self) -> bool:
-        return self.gpus_required is not None or self.gpu_memory_required is not None
+    async def run_job(
+        self, resources_required: Optional[ResourcesInternal], job: Job
+    ) -> JobCompletion[Any]:
+        if resources_required is None:
+            raise ValueError(
+                "Resources.logical_cpu and memory_gb must be specified for "
+                "AllocCloudInstance"
+            )
 
-    def needs_cuda(self) -> bool:
-        return self.uses_gpu() and _needs_cuda(self.flags_required)
-
-    async def run_job(self, job: Job) -> JobCompletion[Any]:
-        resources_required = Resources.from_cpu_and_memory(
-            self.logical_cpu_required,
-            self.memory_gb_required,
-            self.interruption_probability_threshold,
-            self.gpus_required,
-            self.gpu_memory_required,
-            self.flags_required,
-        )
         if self.cloud_provider == "EC2":
             return await run_job_ec2_instance_registrar(
                 job,
@@ -894,39 +1009,87 @@ class AllocCloudInstance(Host):
                 f"Unexpected value for cloud_provider {self.cloud_provider}"
             )
 
+    async def run_map(
+        self,
+        function: Callable[[_T], _U],
+        args: Sequence[_T],
+        resources_required_per_task: Optional[ResourcesInternal],
+        job_fields: Dict[str, Any],
+        num_concurrent_tasks: int,
+        pickle_protocol: int,
+    ) -> Sequence[_U]:
+        if resources_required_per_task is None:
+            raise ValueError(
+                "Resources.logical_cpu and memory_gb must be specified for "
+                "AllocCloudInstance"
+            )
 
-@dataclasses.dataclass(frozen=True)
-class AllocCloudInstances:
-    """
-    Specifies the requirements for a set of cloud instances (e.g. EC2 instances or Azure
-    VMs) for running multiple workers. Also see
-    [AllocCloudInstance][meadowrun.AllocCloudInstance]
+        if self.cloud_provider == "EC2":
+            helper = await prepare_ec2_run_map(
+                function,
+                args,
+                self.region_name,
+                resources_required_per_task,
+                num_concurrent_tasks,
+                job_fields["ports"],
+            )
+        elif self.cloud_provider == "AzureVM":
+            helper = await prepare_azure_vm_run_map(
+                function,
+                args,
+                self.region_name,
+                resources_required_per_task,
+                num_concurrent_tasks,
+                job_fields["ports"],
+            )
+        else:
+            raise ValueError(
+                f"Unexpected value for cloud_provider {self.cloud_provider}"
+            )
 
-    Attributes:
-        logical_cpu_required_per_task:
-        memory_gb_required_per_task:
-        interruption_probability_threshold: See
-            [AllocCloudInstance][meadowrun.AllocCloudInstance]
-        cloud_provider: Either `EC2` or `AzureVM`
-        gpus_required_per_task:
-        gpu_memory_required_per_task: Total GPU memory required across all GPUs
-        flags_required: e.g. "intel", "avx512", etc.
-        num_concurrent_tasks: The number of workers to launch. In the context of a
-            [run_map][meadowrun.run_map] call, this can be less than or equal to the
-            number of args/tasks. Will default to half the total number of tasks if set
-            to None.
-        region_name:
-    """
+        # Now we will run worker_loop jobs on the hosts we got:
 
-    logical_cpu_required_per_task: float
-    memory_gb_required_per_task: float
-    interruption_probability_threshold: float
-    cloud_provider: CloudProviderType
-    gpus_required_per_task: Optional[float] = None
-    gpu_memory_required_per_task: Optional[float] = None
-    flags_required: Union[Iterable[str], str, None] = None
-    num_concurrent_tasks: Optional[int] = None
-    region_name: Optional[str] = None
+        pickled_worker_function = cloudpickle.dumps(
+            helper.worker_function, protocol=pickle_protocol
+        )
+
+        worker_tasks = []
+        worker_id = 0
+        for public_address, worker_job_ids in helper.allocated_hosts.items():
+            for worker_job_id in worker_job_ids:
+                job = Job(
+                    job_id=worker_job_id,
+                    py_function=PyFunctionJob(
+                        pickled_function=pickled_worker_function,
+                        pickled_function_arguments=pickle.dumps(
+                            ([public_address, worker_id], {}), protocol=pickle_protocol
+                        ),
+                    ),
+                    **job_fields,
+                )
+
+                worker_tasks.append(
+                    asyncio.create_task(
+                        SshHost(
+                            public_address,
+                            helper.ssh_username,
+                            helper.ssh_private_key,
+                            (self.cloud_provider, helper.region_name),
+                        ).run_job(resources_required_per_task, job)
+                    )
+                )
+
+                worker_id += 1
+
+        # finally, wait for results:
+
+        results = await helper.results_future
+
+        # TODO if there's an error these workers will crash before the results_future
+        # returns
+        await asyncio.gather(*worker_tasks)
+
+        return results
 
 
 def _pickle_protocol_for_deployed_interpreter() -> int:
@@ -980,12 +1143,68 @@ def _prepare_ports(
         return [str(p) for p in ports]
 
 
+def _prepare_sidecar_containers(
+    container_interpreters: Union[
+        Iterable[ContainerInterpreterBase], ContainerInterpreterBase, None
+    ] = None,
+) -> Tuple[Sequence[ContainerImage], Sequence[CredentialsSourceMessage]]:
+    if container_interpreters is None:
+        return [], []
+
+    if isinstance(container_interpreters, ContainerInterpreterBase):
+        container_interpreters = [container_interpreters]
+
+    sidecar_containers = []
+    credentials_sources = []
+    for interpreter in container_interpreters:
+        if isinstance(interpreter, ContainerAtDigestInterpreter):
+            sidecar_containers.append(
+                ContainerImage(
+                    container_image_at_digest=ContainerAtDigest(
+                        repository=interpreter.repository_name,
+                        digest=interpreter.digest,
+                    )
+                )
+            )
+        elif isinstance(interpreter, ContainerInterpreter):
+            sidecar_containers.append(
+                ContainerImage(
+                    container_image_at_tag=ContainerAtTag(
+                        repository=interpreter.repository_name,
+                        tag=interpreter.tag,
+                    )
+                )
+            )
+        else:
+            raise ValueError(
+                f"Unexpected type of sidecar_container {type(interpreter)}"
+            )
+
+        username_password_secret = interpreter._get_username_password_secret()
+        if username_password_secret is not None:
+            credentials_sources.append(
+                _credentials_source_message(
+                    username_password_secret._to_credentials_source(
+                        "DOCKER",
+                        get_registry_domain(interpreter._get_repository_name())[0],
+                        Credentials.Type.USERNAME_PASSWORD,
+                    )
+                )
+            )
+
+    return sidecar_containers, credentials_sources
+
+
 async def run_function(
     function: Union[Callable[..., _T], str],
     host: Host,
+    resources: Optional[Resources] = None,
     deployment: Optional[Deployment] = None,
     args: Optional[Sequence[Any]] = None,
     kwargs: Optional[Dict[str, Any]] = None,
+    sidecar_containers: Union[
+        Iterable[ContainerInterpreterBase], ContainerInterpreterBase, None
+    ] = None,
     ports: Union[Iterable[str], str, Iterable[int], int, None] = None,
 ) -> _T:
     """
@@ -996,12 +1215,17 @@ async def run_function(
             lambda, or a string like `"package.module.function_name"` (which is useful
             if the function cannot be referenced in the current environment but can be
             referenced in the deployed environment)
-        host: See [AllocCloudInstance][meadowrun.AllocCloudInstance]. Specifies what
-            resources are needed to run this function
+        host: Specifies where to run the function. See
+            [AllocCloudInstance][meadowrun.AllocCloudInstance].
+        resources: Specifies the resources (e.g. CPU, RAM) needed by the
+            function. For some hosts, this is optional, for other hosts it is required.
+            See [Resources][meadowrun.Resources].
         deployment: See [Deployment][meadowrun.Deployment]. Specifies the
             environment (code and libraries) that are needed to run this function
         args: Passed to the function like `function(*args)`
         kwargs: Passed to the function like `function(**kwargs)`
+        sidecar_containers: Additional containers that will be available from the main
+            job as sidecar-container-0 sidecar-container-1, etc.
         ports: A specification of ports to make available on the machine that runs this
             job. E.g. 8000, "8080-8089" (inclusive). Ports will be opened just for the
             duration of this job. Be careful as other jobs could be running on the same
@@ -1011,6 +1235,9 @@ async def run_function(
         The result of calling `function`
     """
 
+    if resources is None:
+        resources = Resources()
+
     pickle_protocol = _pickle_protocol_for_deployed_interpreter()
 
     # first pickle the function arguments from job_run_spec
@@ -1062,7 +1289,7 @@ async def run_function(
         environment_variables,
         credentials_sources,
     ) = _add_defaults_to_deployment(deployment)
-    if host.needs_cuda() and isinstance(
+    if resources.needs_cuda() and isinstance(
         interpreter, (EnvironmentSpec, EnvironmentSpecInCode)
     ):
         interpreter.additional_software["cuda"] = ""
@@ -1071,27 +1298,41 @@ async def run_function(
         code, _DeploymentTarget.from_single_host(host)
     )
 
+    (
+        sidecar_containers_prepared,
+        additional_credentials_sources,
+    ) = _prepare_sidecar_containers(sidecar_containers)
+    credentials_sources.extend(additional_credentials_sources)
+
     job = Job(
         job_id=job_id,
         job_friendly_name=friendly_name,
         environment_variables=environment_variables,
         result_highest_pickle_protocol=pickle.HIGHEST_PROTOCOL,
         py_function=py_function,
+        sidecar_containers=sidecar_containers_prepared,
         credentials_sources=credentials_sources,
         ports=_prepare_ports(ports),
-        uses_gpu=host.uses_gpu(),
+        uses_gpu=resources.uses_gpu(),
+        **{  # type: ignore
+            _job_field_for_code_deployment(code): code,
+            _job_field_for_interpreter_deployment(interpreter): interpreter,
+        },
     )
-    _add_deployments_to_job(job, code, interpreter)
 
-    job_completion = await host.run_job(job)
+    job_completion = await host.run_job(resources.to_internal(), job)
     return job_completion.result
 
 
 async def run_command(
     args: Union[str, Sequence[str]],
     host: Host,
+    resources: Optional[Resources] = None,
     deployment: Optional[Deployment] = None,
     context_variables: Optional[Dict[str, Any]] = None,
+    sidecar_containers: Union[
+        Iterable[ContainerInterpreterBase], ContainerInterpreterBase, None
+    ] = None,
     ports: Union[Iterable[str], str, Iterable[int], int, None] = None,
 ) -> JobCompletion[None]:
     """
@@ -1101,11 +1342,16 @@ async def run_command(
         args: Specifies the command to run, can be a string (e.g. `"jupyter nbconvert
             --to html analysis.ipynb"`) or a list of strings (e.g. `["jupyter",
             --"nbconvert", "--to", "html", "analysis.ipynb"]`)
-        host: See [AllocCloudInstance][meadowrun.AllocCloudInstance]. Specifies what
-            resources are needed to run this command
+        host: Specifies where to run the function. See
+            [AllocCloudInstance][meadowrun.AllocCloudInstance].
+        resources: Specifies the resources (e.g. CPU, RAM) needed by the
+            command. For some hosts, this is optional, for other hosts it is required.
+            See [Resources][meadowrun.Resources].
         deployment: See [Deployment][meadowrun.Deployment]. Specifies
             the environment (code and libraries) that are needed to run this command
         context_variables: Experimental feature
+        sidecar_containers: Additional containers that will be available from the main
+            job as sidecar-container-0 sidecar-container-1, etc.
         ports: A specification of ports to make available on the machine that runs this
             job. E.g. 8000, "8080-8089" (inclusive). Ports will be opened just for the
             duration of this job. Be careful as other jobs could be running on the same
@@ -1115,6 +1361,9 @@ async def run_command(
         A JobCompletion object that contains metadata about the running of the job.
     """
 
+    if resources is None:
+        resources = Resources()
+
     job_id = str(uuid.uuid4())
     if isinstance(args, str):
         args = shlex.split(args)
@@ -1129,7 +1378,7 @@ async def run_command(
         credentials_sources,
     ) = _add_defaults_to_deployment(deployment)
 
-    if host.needs_cuda() and isinstance(
+    if resources.needs_cuda() and isinstance(
         interpreter, (EnvironmentSpec, EnvironmentSpecInCode)
     ):
         interpreter.additional_software["cuda"] = ""
@@ -1145,6 +1394,12 @@ async def run_command(
     else:
         pickled_context_variables = b""
 
+    (
+        sidecar_containers_prepared,
+        additional_credentials_sources,
+    ) = _prepare_sidecar_containers(sidecar_containers)
+    credentials_sources.extend(additional_credentials_sources)
+
     job = Job(
         job_id=job_id,
         job_friendly_name=_make_valid_friendly_name(friendly_name),
@@ -1153,20 +1408,29 @@ async def run_command(
         py_command=PyCommandJob(
             command_line=args, pickled_context_variables=pickled_context_variables
         ),
+        sidecar_containers=sidecar_containers_prepared,
         credentials_sources=credentials_sources,
         ports=_prepare_ports(ports),
-        uses_gpu=host.uses_gpu(),
+        uses_gpu=resources.uses_gpu(),
+        **{  # type: ignore
+            _job_field_for_code_deployment(code): code,
+            _job_field_for_interpreter_deployment(interpreter): interpreter,
+        },
     )
-    _add_deployments_to_job(job, code, interpreter)
 
-    return await host.run_job(job)
+    return await host.run_job(resources.to_internal(), job)
 
 
 async def run_map(
     function: Callable[[_T], _U],
     args: Sequence[_T],
-    hosts: AllocCloudInstances,
+    host: Host,
+    resources_per_task: Optional[Resources] = None,
     deployment: Optional[Deployment] = None,
+    num_concurrent_tasks: Optional[int] = None,
+    sidecar_containers: Union[
+        Iterable[ContainerInterpreterBase], ContainerInterpreterBase, None
+    ] = None,
     ports: Union[Iterable[str], str, Iterable[int], int, None] = None,
 ) -> Sequence[_U]:
     """
@@ -1177,10 +1441,18 @@ async def run_map(
             lambda
         args: A list of objects, each item in the list represents a "task",
             where each "task" is an invocation of `function` on the item in the list
-        hosts: See [AllocCloudInstances][meadowrun.AllocCloudInstances]. Specifies how
-            many workers to provision and what resources are needed for each worker.
+        resources_per_task: The resources (e.g. CPU and RAM) required to run a
+            single task. For some hosts, this is optional, for other hosts it is
+            required. See [Resources][meadowrun.Resources].
+        host: Specifies where to get compute resources from. See
+            [AllocCloudInstance][meadowrun.AllocCloudInstance].
+        num_concurrent_tasks: The number of workers to launch. This can be less than or
+            equal to the number of args/tasks. Will default to half the total number of
+            tasks plus one, rounded down if set to None.
         deployment: See [Deployment][meadowrun.Deployment]. Specifies the environment
             (code and libraries) that are needed to run the function
+        sidecar_containers: Additional containers that will be available from the main
+            job as sidecar-container-0 sidecar-container-1, etc.
         ports: A specification of ports to make available on the machines that runs
             tasks for this job. E.g. 8000, "8080-8089" (inclusive). Ports will be opened
             just for the duration of this job. Be careful as other jobs could be running
@@ -1190,42 +1462,13 @@ async def run_map(
         Returns the result of running `function` on each of `args`
     """
 
-    if not hosts.num_concurrent_tasks:
+    if resources_per_task is None:
+        resources_per_task = Resources()
+
+    if not num_concurrent_tasks:
         num_concurrent_tasks = len(args) // 2 + 1
     else:
-        num_concurrent_tasks = min(hosts.num_concurrent_tasks, len(args))
-
-    resources_required_per_task = Resources.from_cpu_and_memory(
-        hosts.logical_cpu_required_per_task,
-        hosts.memory_gb_required_per_task,
-        hosts.interruption_probability_threshold,
-        hosts.gpus_required_per_task,
-        hosts.gpu_memory_required_per_task,
-        hosts.flags_required,
-    )
-
-    prepared_ports = _prepare_ports(ports)
-
-    if hosts.cloud_provider == "EC2":
-        helper = await prepare_ec2_run_map(
-            function,
-            args,
-            hosts.region_name,
-            resources_required_per_task,
-            num_concurrent_tasks,
-            prepared_ports,
-        )
-    elif hosts.cloud_provider == "AzureVM":
-        helper = await prepare_azure_vm_run_map(
-            function,
-            args,
-            hosts.region_name,
-            resources_required_per_task,
-            num_concurrent_tasks,
-            prepared_ports,
-        )
-    else:
-        raise ValueError(f"Unexpected value for cloud_provider {hosts.cloud_provider}")
+        num_concurrent_tasks = min(num_concurrent_tasks, len(args))
 
     # prepare some variables for constructing the worker jobs
     friendly_name = _get_friendly_name(function)
@@ -1236,106 +1479,83 @@ async def run_map(
         credentials_sources,
     ) = _add_defaults_to_deployment(deployment)
 
-    uses_gpu = (
-        hosts.gpus_required_per_task is not None
-        or hosts.gpu_memory_required_per_task is not None
-    )
-    if (
-        uses_gpu
-        and _needs_cuda(hosts.flags_required)
-        and isinstance(interpreter, (EnvironmentSpec, EnvironmentSpecInCode))
+    if resources_per_task.needs_cuda() and isinstance(
+        interpreter, (EnvironmentSpec, EnvironmentSpecInCode)
     ):
         interpreter.additional_software["cuda"] = ""
 
-    code = await _prepare_code_deployment(code, _DeploymentTarget.from_hosts(hosts))
+    code = await _prepare_code_deployment(
+        code, _DeploymentTarget.from_single_host(host)
+    )
 
     pickle_protocol = _pickle_protocol_for_deployed_interpreter()
 
-    # Now we will run worker_loop jobs on the hosts we got:
+    (
+        sidecar_containers_prepared,
+        additional_credentials_sources,
+    ) = _prepare_sidecar_containers(sidecar_containers)
+    credentials_sources.extend(additional_credentials_sources)
 
-    pickled_worker_function = cloudpickle.dumps(
-        helper.worker_function, protocol=pickle_protocol
+    prepared_ports = _prepare_ports(ports)
+
+    job_fields = {
+        "job_friendly_name": friendly_name,
+        "environment_variables": environment_variables,
+        "result_highest_pickle_protocol": pickle.HIGHEST_PROTOCOL,
+        "sidecar_containers": sidecar_containers_prepared,
+        "credentials_sources": credentials_sources,
+        "ports": prepared_ports,
+        "uses_gpu": resources_per_task.uses_gpu(),
+        _job_field_for_code_deployment(code): code,
+        _job_field_for_interpreter_deployment(interpreter): interpreter,
+    }
+
+    return await host.run_map(
+        function,
+        args,
+        resources_per_task.to_internal(),
+        job_fields,
+        num_concurrent_tasks,
+        pickle_protocol,
     )
 
-    worker_tasks = []
-    worker_id = 0
-    for public_address, worker_job_ids in helper.allocated_hosts.items():
-        for worker_job_id in worker_job_ids:
-            job = Job(
-                job_id=worker_job_id,
-                job_friendly_name=friendly_name,
-                environment_variables=environment_variables,
-                result_highest_pickle_protocol=pickle.HIGHEST_PROTOCOL,
-                py_function=PyFunctionJob(
-                    pickled_function=pickled_worker_function,
-                    pickled_function_arguments=pickle.dumps(
-                        ([public_address, worker_id], {}), protocol=pickle_protocol
-                    ),
-                ),
-                credentials_sources=credentials_sources,
-                ports=prepared_ports,
-                uses_gpu=uses_gpu,
-            )
-            _add_deployments_to_job(job, code, interpreter)
 
-            worker_tasks.append(
-                asyncio.create_task(
-                    SshHost(
-                        public_address,
-                        helper.ssh_username,
-                        helper.ssh_private_key,
-                        (hosts.cloud_provider, helper.region_name),
-                    ).run_job(job)
-                )
-            )
-
-            worker_id += 1
-
-    # finally, wait for results:
-
-    results = await helper.results_future
-
-    # TODO if there's an error these workers will crash before the results_future
-    # returns
-    await asyncio.gather(*worker_tasks)
-
-    return results
-
-
-def _add_deployments_to_job(
-    job: Job,
+def _job_field_for_code_deployment(
     code_deployment: Union[CodeDeployment, VersionedCodeDeployment],
-    interpreter_deployment: Union[
-        InterpreterDeployment, VersionedInterpreterDeployment
-    ],
-) -> None:
+) -> str:
     """
-    Think of this as job.code_deployment = code_deployment; job.interpreter_deployment =
-    interpreter_deployment, but it's complicated because these are protobuf oneofs
+    You can't just do Job(code_deployment=code_deployment) because of the protobuf
+    oneofs. Instead you need to do
+    Job(**{_job_field_for_code_deployment(code_deployment): code_deployment})
     """
     if isinstance(code_deployment, ServerAvailableFolder):
-        job.server_available_folder.CopyFrom(code_deployment)
+        return "server_available_folder"
     elif isinstance(code_deployment, GitRepoCommit):
-        job.git_repo_commit.CopyFrom(code_deployment)
+        return "git_repo_commit"
     elif isinstance(code_deployment, GitRepoBranch):
-        job.git_repo_branch.CopyFrom(code_deployment)
+        return "git_repo_branch"
     elif isinstance(code_deployment, CodeZipFile):
-        job.code_zip_file.CopyFrom(code_deployment)
+        return "code_zip_file"
     else:
         raise ValueError(f"Unknown code deployment type {type(code_deployment)}")
 
+
+def _job_field_for_interpreter_deployment(
+    interpreter_deployment: Union[InterpreterDeployment, VersionedInterpreterDeployment]
+) -> str:
+    """See _job_field_for_code_deployment for usage pattern"""
     if isinstance(interpreter_deployment, ServerAvailableInterpreter):
-        job.server_available_interpreter.CopyFrom(interpreter_deployment)
+        return "server_available_interpreter"
     elif isinstance(interpreter_deployment, ContainerAtDigest):
-        job.container_at_digest.CopyFrom(interpreter_deployment)
+        return "container_at_digest"
     elif isinstance(interpreter_deployment, ServerAvailableContainer):
-        job.server_available_container.CopyFrom(interpreter_deployment)
+        return "server_available_container"
     elif isinstance(interpreter_deployment, ContainerAtTag):
-        job.container_at_tag.CopyFrom(interpreter_deployment)
+        return "container_at_tag"
     elif isinstance(interpreter_deployment, EnvironmentSpecInCode):
-        job.environment_spec_in_code.CopyFrom(interpreter_deployment)
+        return "environment_spec_in_code"
     elif isinstance(interpreter_deployment, EnvironmentSpec):
-        job.environment_spec.CopyFrom(interpreter_deployment)
+        return "environment_spec"
     else:
         raise ValueError(
             f"Unknown interpreter deployment type {type(interpreter_deployment)}"
