@@ -8,15 +8,17 @@ import asyncio
 import dataclasses
 import pickle
 from typing import (
-    TYPE_CHECKING,
     Any,
     Awaitable,
     Callable,
     Coroutine,
     Dict,
     Generic,
+    Iterable,
     List,
     Optional,
+    Sequence,
+    TYPE_CHECKING,
     Tuple,
     Type,
     TypeVar,
@@ -30,10 +32,12 @@ import meadowrun.ssh as ssh
 
 if TYPE_CHECKING:
     from meadowrun.credentials import UsernamePassword
-    from meadowrun.instance_selection import Resources
+from meadowrun.instance_selection import ResourcesInternal
 from meadowrun.meadowrun_pb2 import Job, ProcessState
 
+
 _T = TypeVar("_T")
+_U = TypeVar("_U")
 
 
 CloudProvider = "EC2", "AzureVM"
@@ -45,6 +49,7 @@ async def _retry(
     exception_types: Union[Type, Tuple[Type, ...]],
     max_num_attempts: int = 5,
     delay_seconds: float = 1,
+    message: str = "Retrying on error",
 ) -> _T:
     i = 0
     while True:
@@ -55,21 +60,89 @@ async def _retry(
             if i >= max_num_attempts:
                 raise
             else:
-                print(f"Retrying on error: {e}")
+                print(f"{message}: {e}")
                 await asyncio.sleep(delay_seconds)
 
 
-class Host(abc.ABC):
-    @abc.abstractmethod
+def _needs_cuda(flags_required: Union[Iterable[str], str, None]) -> bool:
+    if flags_required is None:
+        return False
+    if isinstance(flags_required, str):
+        return "nvidia" == flags_required
+    return "nvidia" in flags_required
+
+
+@dataclasses.dataclass(frozen=True)
+class Resources:
+    """
+    Specifies the requirements for a job or for each task within a job
+
+    Attributes:
+        logical_cpu: Specifies logical CPU (aka vCPU) required. E.g. 2 means we require
+            2 logical CPUs
+        memory_gb: Specifies RAM required. E.g. 1.5 means we requires 1.5 GB of RAM
+        max_eviction_rate: Specifies what eviction rate (aka interruption probability)
+            we're okay with as a percent. E.g. `80` means that any instance type with an
+            interruption probability less than 80% can be used. Use `0` to indicate that
+            only on-demand instance are acceptable (i.e. do not use spot instances)
+        gpus: Number of GPUs required. If gpu_memory is set, but this value is not set,
+            this is implied to be 1
+        gpu_memory: Total GPU memory (aka VRAM) required across all GPUs
+        flags_required: E.g. "intel", "avx512", etc.
+    """
+
+    logical_cpu: Optional[float] = None
+    memory_gb: Optional[float] = None
+    max_eviction_rate: float = 80
+    gpus: Optional[float] = None
+    gpu_memory: Optional[float] = None
+    flags: Union[Iterable[str], str, None] = None
+
     def uses_gpu(self) -> bool:
-        pass
+        return self.gpus is not None or self.gpu_memory is not None
 
-    @abc.abstractmethod
     def needs_cuda(self) -> bool:
+        return self.uses_gpu() and _needs_cuda(self.flags)
+
+    def to_internal(self) -> Optional[ResourcesInternal]:
+        if self.logical_cpu is None or self.memory_gb is None:
+            return None
+        else:
+            return ResourcesInternal.from_cpu_and_memory(
+                self.logical_cpu,
+                self.memory_gb,
+                self.max_eviction_rate,
+                self.gpus,
+                self.gpu_memory,
+                self.flags,
+            )
+
+
+class Host(abc.ABC):
+    """
+    Host is an abstract class for specifying where to run a job. See implementations
+    below.
+    """
+
+    @abc.abstractmethod
+    async def run_job(
+        self, resources_required: Optional[ResourcesInternal], job: Job
+    ) -> JobCompletion[Any]:
         pass
 
     @abc.abstractmethod
-    async def run_job(self, job: Job) -> JobCompletion[Any]:
+    async def run_map(
+        self,
+        function: Callable[[_T], _U],
+        args: Sequence[_T],
+        resources_required_per_task: Optional[ResourcesInternal],
+        job_fields: Dict[str, Any],
+        num_concurrent_tasks: int,
+        pickle_protocol: int,
+    ) -> Sequence[_U]:
+        # Note for implementors: job_fields will be populated with everything other than
+        # job_id and py_function, so the implementation should construct
+        # Job(job_id=job_id, py_function=py_function, **job_fields)
         pass
 
 
@@ -89,13 +162,9 @@ class SshHost(Host):
     # the InstanceRegistrar that we used to allocate this job is.
     cloud_provider: Optional[Tuple[CloudProviderType, str]] = None
 
-    def uses_gpu(self) -> bool:
-        return False
-
-    def needs_cuda(self) -> bool:
-        return False
-
-    async def run_job(self, job: Job) -> JobCompletion[Any]:
+    async def run_job(
+        self, resources_required: Optional[ResourcesInternal], job: Job
+    ) -> JobCompletion[Any]:
         # try the connection 20 times.
         connection = await _retry(
             lambda: ssh.connect(
@@ -192,6 +261,8 @@ class SshHost(Host):
                 remote_paths = " ".join(
                     [
                         f"{job_io_prefix}.job_to_run",
+                        f"{job_io_prefix}.function",
+                        f"{job_io_prefix}.arguments",
                         f"{job_io_prefix}.state",
                         f"{job_io_prefix}.result",
                         f"{job_io_prefix}.process_state",
@@ -204,6 +275,8 @@ class SshHost(Host):
                     await ssh.run_and_capture(
                         connection, f"rm -f {remote_paths}", check=True
                     )
+                except asyncio.CancelledError:
+                    raise
                 except Exception as e:
                     print(
                         f"Error cleaning up files on remote machine: "
@@ -212,14 +285,16 @@ class SshHost(Host):
             connection.close()
             await connection.wait_closed()
 
-
-@dataclasses.dataclass(frozen=True)
-class AllocCloudInstancesInternal:
-    """Identical to AllocCloudInstances but all values must be set"""
-
-    resources_required_per_task: Resources
-    num_concurrent_tasks: int
-    region_name: str
+    async def run_map(
+        self,
+        function: Callable[[_T], _U],
+        args: Sequence[_T],
+        resources_required_per_task: Optional[ResourcesInternal],
+        job_fields: Dict[str, Any],
+        num_concurrent_tasks: int,
+        pickle_protocol: int,
+    ) -> Sequence[_U]:
+        raise NotImplementedError("run_map is not implemented for SshHost")
 
 
 @dataclasses.dataclass
