@@ -12,22 +12,28 @@ import shutil
 import sys
 import traceback
 from typing import (
-    TYPE_CHECKING,
     Any,
+    Callable,
     Coroutine,
     Dict,
     Iterable,
     List,
     Optional,
     Sequence,
+    TYPE_CHECKING,
     Tuple,
+    TypeVar,
 )
 
+from meadowrun.aws_integration.ecr import get_ecr_username_password
+from meadowrun.azure_integration.acr import get_acr_username_password
+
 if TYPE_CHECKING:
     from typing_extensions import Literal
 
     from meadowrun._vendor import aiodocker
     from meadowrun._vendor.aiodocker import containers as aiodocker_containers
+    from meadowrun.instance_selection import ResourcesInternal
 
 from meadowrun.config import (
     MEADOWRUN_AGENT_PID,
@@ -46,10 +52,11 @@ from meadowrun.deployment_manager import (
     get_code_paths,
 )
 from meadowrun.docker_controller import (
-    run_container,
     get_image_environment_variables_and_working_dir,
+    get_registry_domain,
     pull_image,
     remove_container,
+    run_container,
 )
 from meadowrun.meadowrun_pb2 import (
     Credentials,
@@ -68,6 +75,9 @@ from meadowrun.shared import pickle_exception
 
 ProcessStateEnum = ProcessState.ProcessStateEnum
 
+_T = TypeVar("_T")
+_U = TypeVar("_U")
+
 
 _MEADOWRUN_CONTEXT_VARIABLES = "MEADOWRUN_CONTEXT_VARIABLES"
 _MEADOWRUN_RESULT_FILE = "MEADOWRUN_RESULT_FILE"
@@ -100,10 +110,6 @@ class _JobSpecTransformed:
     # expose files on the host machine for input/output with the container.
     container_binds: List[Tuple[str, str]]
 
-    ports: List[str]
-
-    uses_gpu: bool
-
     environment_variables: Dict[str, str] = dataclasses.field(
         default_factory=lambda: {}
     )
@@ -175,8 +181,6 @@ def _prepare_py_command(
         # we need a list, not a protobuf fake list
         list(job.py_command.command_line),
         _io_file_container_binds(io_folder, io_files),
-        list(job.ports),
-        job.uses_gpu,
         environment,
     )
 
@@ -290,8 +294,6 @@ def _prepare_py_function(
             itertools.chain(io_files, io_files_for_function, io_files_for_arguments),
         )
         + [(_FUNC_WORKER_PATH, func_worker_path)],
-        list(job.ports),
-        job.uses_gpu,
     )
 
 
@@ -299,6 +301,7 @@ async def _launch_non_container_job(
     job_spec_type: Literal["py_command", "py_function"],
     job_spec_transformed: _JobSpecTransformed,
     code_paths: Sequence[str],
+    cwd_path: Optional[str],
     log_file_name: str,
     job: Job,
     io_folder: str,
@@ -350,8 +353,8 @@ async def _launch_non_container_job(
     # (3) get the working directory and fix the command line
 
     paths_to_search = env_vars["PATH"]
-    if code_paths:
-        working_directory = code_paths[0]
+    if cwd_path:
+        working_directory = cwd_path
         paths_to_search = f"{working_directory};{paths_to_search}"
     else:
         # TODO probably cleanest to allocate a new working directory for each job
@@ -446,9 +449,12 @@ async def _launch_container_job(
     container_image_name: str,
     job_spec_transformed: _JobSpecTransformed,
     code_paths: Sequence[str],
+    cwd_path: Optional[str],
     log_file_name: str,
     job: Job,
     io_folder: str,
+    machine_cache_folder: str,
+    sidecar_container_images: List[str],
 ) -> Tuple[str, Coroutine[Any, Any, ProcessState]]:
     """
     Contains logic specific to launching jobs that run in a container. Only separated
@@ -463,17 +469,21 @@ async def _launch_container_job(
     continuation.
     """
 
-    os.makedirs("/var/meadowrun/machine_cache", exist_ok=True)
+    binds: List[Tuple[str, str]] = [(machine_cache_folder, "/meadowrun/machine_cache")]
 
-    binds: List[Tuple[str, str]] = [
-        ("/var/meadowrun/machine_cache", "/meadowrun/machine_cache")
-    ]
-    mounted_code_paths: List[str] = []
+    if cwd_path is not None:
+        unique_code_paths = list(dict.fromkeys(itertools.chain([cwd_path], code_paths)))
+    else:
+        unique_code_paths = list(code_paths)
 
     # populate binds with code paths we need
-    for i, path_on_host in enumerate(code_paths):
+    # TODO this isn't exactly right--some code paths "overlap", like foo/a and foo/a/b.
+    # We should just bind foo/a once as e.g. /meadowrun/code0, and then add both
+    # /meadowrun/code0 and /meadowrun/code0/b to the PYTHONPATH. Instead, right now, we
+    # mount /foo/a as /meadowrun/code0 and /foo/a/b as /meadowrun/code1, which might
+    # lead to unexpected behavior.
+    for i, path_on_host in enumerate(unique_code_paths):
         mounted_code_path = f"{MEADOWRUN_CODE_MOUNT_LINUX}{i}"
-        mounted_code_paths.append(mounted_code_path)
         binds.append((path_on_host, mounted_code_path))
 
     # get the image's environment variables and working directory
@@ -485,9 +495,11 @@ async def _launch_container_job(
     # If the container image has a meaningful working directory set, then we want to
     # leave that as it is. But if the working directory is "/tmp/", we use that as a
     # placeholder to mean that we don't care about the container image's working
-    # directory. In that case, we set it to the first mounted code path.
-    if working_dir in ("/tmp/", "/tmp") and len(mounted_code_paths) > 0:
-        working_dir = mounted_code_paths[0]
+    # directory. In that case, we set it to cwd_path, if given.
+    if working_dir in ("/tmp/", "/tmp") and cwd_path is not None:
+        (working_dir,) = tuple(
+            mounted_path for (host_path, mounted_path) in binds if host_path == cwd_path
+        )
 
     # If we've exposed any code paths, add them to PYTHONPATH. The normal behavior for
     # environment variables is that if they're specified in job_spec_transformed, those
@@ -495,7 +507,7 @@ async def _launch_container_job(
     # specified in job_spec_transformed, we use whatever is specified in the container
     # image. We need to replicate that logic here so that we just add mounted_code_paths
     # to whatever PYTHONPATH would have "normally" become.
-    if mounted_code_paths:
+    if len(binds) > 1:
         existing_python_path = []
         if "PYTHONPATH" in job_spec_transformed.environment_variables:
             existing_python_path = [
@@ -511,14 +523,41 @@ async def _launch_container_job(
                         # environment variables
                         break
 
+        code_paths = tuple(
+            mounted_path
+            for (host_path, mounted_path) in binds
+            if host_path in code_paths
+        )
         # TODO we need to use ":" for Linux and ";" for Windows containers
         job_spec_transformed.environment_variables["PYTHONPATH"] = ":".join(
-            itertools.chain(mounted_code_paths, existing_python_path)
+            itertools.chain(code_paths, existing_python_path)
         )
 
     # now, expose any files we need for communication with the container
     binds.extend(job_spec_transformed.container_binds)
 
+    # now run any sidecar_containers that were specified
+    docker_client = None
+
+    sidecar_container_ips = []
+    sidecar_containers = []
+    for sidecar_container_image in sidecar_container_images:
+        sidecar_container, docker_client = await run_container(
+            docker_client,
+            sidecar_container_image,
+            None,
+            {},
+            None,
+            [],
+            [],
+            [],
+            job.uses_gpu,
+        )
+        sidecar_containers.append(sidecar_container)
+        sidecar_container_ips.append(
+            (await sidecar_container.show())["NetworkSettings"]["IPAddress"]
+        )
+
     # finally, run the container
     print(
         f"Running container ({job_spec_type}): "
@@ -526,18 +565,20 @@ async def _launch_container_job(
         f"container image={container_image_name}; "
         f"PYTHONPATH={job_spec_transformed.environment_variables.get('PYTHONPATH')} "
         f"log_file_name={log_file_name}; code paths={','.join(code_paths)} "
-        f"ports={','.join(port for port in job_spec_transformed.ports)}"
+        f"ports={','.join(port for port in job.ports)}"
     )
 
     container, docker_client = await run_container(
+        docker_client,
         container_image_name,
         # json serializer needs a real list, not a protobuf fake list
         job_spec_transformed.command_line,
         job_spec_transformed.environment_variables,
         working_dir,
         binds,
-        job_spec_transformed.ports,
-        job_spec_transformed.uses_gpu,
+        list(job.ports),
+        [(f"sidecar-container-{i}", ip) for i, ip in enumerate(sidecar_container_ips)],
+        job.uses_gpu,
     )
     return container.id, _container_job_continuation(
         container,
@@ -547,6 +588,7 @@ async def _launch_container_job(
         io_folder,
         job.result_highest_pickle_protocol,
         log_file_name,
+        sidecar_containers,
     )
 
 
@@ -558,6 +600,7 @@ async def _container_job_continuation(
     io_folder: str,
     result_highest_pickle_protocol: int,
     log_file_name: str,
+    sidecar_containers: List[aiodocker_containers.DockerContainer],
 ) -> ProcessState:
     """
     Writes the container's logs to log_file_name, waits for the container to finish, and
@@ -601,7 +644,10 @@ async def _container_job_continuation(
         )
     finally:
         try:
-            await remove_container(container)
+            await asyncio.gather(
+                remove_container(container),
+                *(remove_container(c) for c in sidecar_containers),
+            )
         except Exception as e:
             print(f"Warning, unable to remove container: {e}")
         await docker_client.__aexit__(None, None, None)
@@ -685,7 +731,7 @@ def _get_default_working_folder() -> str:
 
 def _set_up_working_folder(
     working_folder: Optional[str],
-) -> Tuple[str, str, str, str, str]:
+) -> Tuple[str, str, str, str, str, str]:
     """
     Sets the working_folder to a default if it's not set, creates the necessary
     subfolders, gets a machine-wide lock on the working folder, then returns io_folder,
@@ -709,12 +755,15 @@ def _set_up_working_folder(
     local_copies_folder = os.path.join(working_folder, "local_copies")
     # misc folder for e.g. storing environment export files sent from local machine
     misc_folder = os.path.join(working_folder, "misc")
+    # machine_cache folder for providing a machine-wide folder on the machine
+    machine_cache_folder = os.path.join(working_folder, "machine_cache")
 
     os.makedirs(io_folder, exist_ok=True)
     os.makedirs(job_logs_folder, exist_ok=True)
     os.makedirs(git_repos_folder, exist_ok=True)
     os.makedirs(local_copies_folder, exist_ok=True)
     os.makedirs(misc_folder, exist_ok=True)
+    os.makedirs(machine_cache_folder, exist_ok=True)
 
     return (
         io_folder,
@@ -722,34 +771,85 @@ def _set_up_working_folder(
         git_repos_folder,
         local_copies_folder,
         misc_folder,
+        machine_cache_folder,
     )
 
 
-async def _get_credentials_for_job(
-    job: Job,
-) -> Tuple[Optional[RawCredentials], Optional[RawCredentials]]:
-    """
-    Gets the credentials for the code_deployment, interpreter_deployment for job_to_run.
-    This is a little silly because we could just have the user specify the
-    code_deployment credentials and the interpreter_deployment credentials explicitly.
-    We should make that change at the same time as we make it possible to register
-    credentials outside of the context of a single job, as we'll need this logic for
-    that use case.
-    """
-
-    # first, get all available credentials sources from the JobToRun
+def _get_credentials_sources(job: Job) -> CredentialsDict:
     credentials_sources: CredentialsDict = {}
     for credentials_source in job.credentials_sources:
         source = credentials_source.WhichOneof("source")
         if source is None:
             raise ValueError(
-                "AddCredentialsRequest request should have a source set: "
+                "CredentialsSourceMessage should have a source set: "
                 f"{credentials_source}"
             )
         credentials_sources.setdefault(credentials_source.service, []).append(
             (credentials_source.service_url, getattr(credentials_source, source))
         )
 
+    return credentials_sources
+
+
+async def _get_credentials_for_docker(
+    repository: str,
+    credentials_sources: CredentialsDict,
+    cloud: Optional[Tuple[CloudProviderType, str]],
+) -> Optional[RawCredentials]:
+    """
+    Tries to get the credentials for a docker repository from credentials_sources. If
+    that doesn't succeed, try to get a username/password for ECR/ACR based on the
+    current role.
+    """
+    result = None
+    try:
+        result = await get_docker_credentials(repository, credentials_sources)
+    except asyncio.CancelledError:
+        raise
+    except Exception:
+        print("Error trying to turn credentials source into actual credentials")
+        traceback.print_exc()
+
+    if result is None and cloud is not None:
+        try:
+            registry_domain, _ = get_registry_domain(repository)
+            if cloud[0] == "EC2":
+                result = get_ecr_username_password(registry_domain)
+            elif cloud[0] == "AzureVM":
+                result = await get_acr_username_password(registry_domain)
+            else:
+                raise ValueError("Unexpected value for CloudProviderType: {cloud[0]}")
+        except asyncio.CancelledError:
+            raise
+        except Exception:
+            print(f"Error trying to get cloud credentials for {repository}")
+            traceback.print_exc()
+
+    return result
+
+
+async def _get_credentials_for_job(
+    job: Job,
+    cloud: Optional[Tuple[CloudProviderType, str]],
+) -> Tuple[
+    Optional[RawCredentials], Optional[RawCredentials], List[Optional[RawCredentials]]
+]:
+    """
+    Returns (credentials for job.code_deployment, credentials for
+    job.interpreter_deployment, credentials for job.sidecar_containers). Credentials for
+    sidecar_containers will be a list of the same length as sidecar_containers, and
+    should be used like zip(job.sidecar_containers, returned_value).
+
+    TODO: This code seems a little convoluted because on the client side, the user
+    specifies which credentials are for which deployment/sidecar container. We should
+    change this code to reflect the user's specifications, but the code we have here is
+    also useful if we want to support registering "global" credentials that are stored
+    outside of the context of a single job.
+    """
+
+    # first, get all available credentials sources from the JobToRun
+    credentials_sources = _get_credentials_sources(job)
+
     # now, get any matching credentials sources and turn them into credentials
     code_deployment_credentials, interpreter_deployment_credentials = None, None
 
@@ -779,17 +879,36 @@ async def _get_credentials_for_job(
         else:
             repository = job.container_at_tag.repository
 
-        try:
-            interpreter_deployment_credentials = await get_docker_credentials(
-                repository, credentials_sources
-            )
-        except asyncio.CancelledError:
-            raise
-        except Exception:
-            print("Error trying to turn credentials source into actual credentials")
-            traceback.print_exc()
+        interpreter_deployment_credentials = await _get_credentials_for_docker(
+            repository, credentials_sources, cloud
+        )
 
-    return code_deployment_credentials, interpreter_deployment_credentials
+    sidecar_container_credentials: List[Optional[RawCredentials]] = []
+    for sidecar_container in job.sidecar_containers:
+        sidecar_container_image = sidecar_container.WhichOneof("container_image")
+
+        if sidecar_container_image in (
+            "container_image_at_digest",
+            "container_image_at_tag",
+        ):
+            if sidecar_container_image == "container_image_at_digest":
+                repository = sidecar_container.container_image_at_digest.repository
+            else:
+                repository = sidecar_container.container_image_at_tag.repository
+
+            sidecar_container_credentials.append(
+                await _get_credentials_for_docker(
+                    repository, credentials_sources, cloud
+                )
+            )
+        else:
+            sidecar_container_credentials.append(None)
+
+    return (
+        code_deployment_credentials,
+        interpreter_deployment_credentials,
+        sidecar_container_credentials,
+    )
 
 
 async def run_local(
@@ -821,17 +940,19 @@ async def run_local(
         git_repos_folder,
         local_copies_folder,
         misc_folder,
+        machine_cache_folder,
     ) = _set_up_working_folder(working_folder)
 
     # unpickle credentials if necessary
     (
         code_deployment_credentials,
         interpreter_deployment_credentials,
-    ) = await _get_credentials_for_job(job)
+        all_sidecar_container_credentials,
+    ) = await _get_credentials_for_job(job, cloud)
 
     try:
         # first, get the code paths
-        code_paths, interpreter_spec_path = await get_code_paths(
+        code_paths, interpreter_spec_path, cwd_path = await get_code_paths(
             git_repos_folder, local_copies_folder, job, code_deployment_credentials
         )
 
@@ -909,10 +1030,16 @@ async def run_local(
         # interpreter
 
         if interpreter_deployment == "server_available_interpreter":
+            if job.sidecar_containers:
+                raise ValueError(
+                    "Cannot specify sidecar_containers with "
+                    "server_available_interpreter"
+                )
             pid, continuation = await _launch_non_container_job(
                 job_spec_type,
                 job_spec_transformed,
                 code_paths,
+                cwd_path,
                 log_file_name,
                 job,
                 io_folder,
@@ -921,14 +1048,19 @@ async def run_local(
             container_id = ""
         elif is_container:
             if interpreter_deployment == "container_at_digest":
-                container_image_name = f"{job.container_at_digest.repository}@{job.container_at_digest.digest}"  # noqa: E501
+                container_image_name = (
+                    f"{job.container_at_digest.repository}"
+                    f"@{job.container_at_digest.digest}"
+                )
                 await pull_image(
                     container_image_name, interpreter_deployment_credentials
                 )
             elif interpreter_deployment == "container_at_tag":
                 # warning this is not reproducible!!! should ideally be resolved on the
                 # client
-                container_image_name = f"{job.container_at_tag.repository}:{job.container_at_tag.tag}"  # noqa: E501
+                container_image_name = (
+                    f"{job.container_at_tag.repository}:{job.container_at_tag.tag}"
+                )
                 await pull_image(
                     container_image_name, interpreter_deployment_credentials
                 )
@@ -941,14 +1073,50 @@ async def run_local(
                     f"Unexpected interpreter_deployment: {interpreter_deployment}"
                 )
 
+            sidecar_containers = []
+            for sidecar_container, sidecar_container_credentials in zip(
+                job.sidecar_containers, all_sidecar_container_credentials
+            ):
+                sidecar_container_image_type = sidecar_container.WhichOneof(
+                    "container_image"
+                )
+                if sidecar_container_image_type == "container_image_at_digest":
+                    sidecar_containers.append(
+                        f"{sidecar_container.container_image_at_digest.repository}"
+                        f"@{sidecar_container.container_image_at_digest.digest}"
+                    )
+                    await pull_image(
+                        sidecar_containers[-1], sidecar_container_credentials
+                    )
+                elif sidecar_container_image_type == "container_image_at_tag":
+                    sidecar_containers.append(
+                        f"{sidecar_container.container_image_at_tag.repository}"
+                        f":{sidecar_container.container_image_at_tag.tag}"
+                    )
+                    await pull_image(
+                        sidecar_containers[-1], sidecar_container_credentials
+                    )
+                elif sidecar_container_image_type == "server_available_container_image":
+                    sidecar_containers.append(
+                        sidecar_container.server_available_container_image.image_name
+                    )
+                else:
+                    raise ValueError(
+                        "Unexpected sidecar container image: "
+                        f"{sidecar_container_image_type}"
+                    )
+
             container_id, continuation = await _launch_container_job(
                 job_spec_type,
                 container_image_name,
                 job_spec_transformed,
                 code_paths,
+                cwd_path,
                 log_file_name,
                 job,
                 io_folder,
+                machine_cache_folder,
+                sidecar_containers,
             )
             # due to the way protobuf works, this is equivalent to None
             pid = 0
@@ -983,13 +1151,12 @@ async def run_local(
 
 @dataclasses.dataclass(frozen=True)
 class LocalHost(Host):
-    def uses_gpu(self) -> bool:
-        return False
+    async def run_job(
+        self, resources_required: Optional[ResourcesInternal], job: Job
+    ) -> JobCompletion[Any]:
+        if resources_required is not None:
+            raise ValueError("Specifying Resources for LocalHost is not supported")
 
-    def needs_cuda(self) -> bool:
-        return False
-
-    async def run_job(self, job: Job) -> JobCompletion[Any]:
         initial_update, continuation = await run_local(job)
         if (
             initial_update.state != ProcessState.ProcessStateEnum.RUNNING
@@ -1017,3 +1184,14 @@ class LocalHost(Host):
             )
         else:
             raise MeadowrunException(result)
+
+    async def run_map(
+        self,
+        function: Callable[[_T], _U],
+        args: Sequence[_T],
+        resources_required_per_task: Optional[ResourcesInternal],
+        job_fields: Dict[str, Any],
+        num_concurrent_tasks: int,
+        pickle_protocol: int,
+    ) -> Sequence[_U]:
+        raise NotImplementedError("run_map on LocalHost is not implemented")
